"""
A server that queues and sends UDP requests for a specific port on a
separate thread.
"""

import queue, socket
from .. util.threads import threads


class Sender:
    def __init__(self, address):
        """
        :param str address: a pair (port, ip_address) to pass to socket.connect
        """
        super().__init__()
        self.address = address
        self.running = True
        self._socket = None
        self.reuse_socket = True

    @property
    def socket(self):
        if not (self.reuse_socket and self._socket):
            self._socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        return self._socket

    def start(self):
        pass

    def stop(self):
        pass

    def send(self, msg):
        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as sock:
            sock.connect(self.address)
            sock.send(msg)
            # sock.sendto(msg, self.address)


class QueuedSender(threads.QueueHandler):
    """
    The UPD protocol is stateless but not necessarily thread-safe.

    ``QueuedSender`` uses a queue to send all UDP messages to one address
    from a new thread.

    """
    def __init__(self, address, **kwds):
        """
        :param str address: a pair (port, ip_address) to pass to socket.connect
        :param bool use_queue: if True, run the connection in a different thread
            with a queue
        """
        super().__init__(**kwds)
        self.sender = Sender(address)

    def send(self, msg):
        self.sender.send(msg)


def sender(address, use_queue=True, **kwds):
    """
    :param tuple address:
    """
    return QueuedSender(address, **kwds) if use_queue else Sender(address)


class Receiver(threads.Loop):
    """
    Receive UDP messages in a thread
    """
    def __init__(self, address, bufsize=0x1000, receive=None, **kwds):
        super().__init__(**kwds)
        self.address = address
        self.bufsize = bufsize
        self.receive = receive or self.receive

    def run(self):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.socket.bind(self.address)

        super().run()

    def loop(self):
        data, addr = self.socket.recvfrom(self.bufsize)
        if data:
            self.receive(data)

    def receive(self, msg):
        pass


class QueuedReceiver(Receiver):
    """
    Receive UDP messages in a thread and put them on a queue.
    """
    def run(self):
        self.queue = queue.Queue()
        super().run()

    def receive(self, msg):
        self.queue.put(msg)



import json, os, sys, yaml

# Allow open to be patched for tests.
open = __builtins__['open']


def dumps(data, **kwds):
    """
    Dumps data into a nicely formatted JSON string.

    :param dict data: a dictionary to dump
    :param kwds: keywords to pass to json.dumps
    :returns: a string with formatted data
    :rtype: str
    """
    return json.dumps(data, indent=4, sort_keys=True, **kwds)


def loads(s, filename=''):
    if filename.endswith('.yml'):
        return yaml.load(s)
    return json.loads(s)


def dump(data, file=sys.stdout, **kwds):
    """
    Dumps data as nicely formatted JSON string to a file or file handle

    :param dict data: a dictionary to dump
    :param file: a filename or file handle to write to
    :param kwds: keywords to pass to json.dump
    """
    def dump(fp):
        json.dump(data, fp, indent=4, sort_keys=True, **kwds)

    if not isinstance(file, str):
        return dump(file)

    if os.path.isabs(file):
        parent = os.path.dirname(file)
        if not os.path.exists(parent):
            os.makedirs(parent, exist_ok=True)

    with open(file, 'w') as fp:
        return dump(fp)


def load(file):
    """
    Loads not only JSON files but also YAML files ending in .yml.

    :param file: a filename or file handle to read from
    :returns: the data loaded from the JSON or YAML file
    :rtype: dict
    """
    if isinstance(file, str):
        fp = open(file)
        filename = file
    else:
        fp = file
        filename = getattr(fp, 'name', '')

    try:
        return loads(fp.read(), filename)

    except Exception as e:
        e.args = ('There was a error in the data file', filename) + e.args
        raise



import queue
from . import (
    attributes, construct, cleanup, defaults, event_queue, load, recurse)
from .. util import exception, json

EVENT_QUEUE_MAXSIZE = 1000


class Project:
    CHILDREN = 'maker', 'drivers', 'layout', 'animation', 'controls'

    @staticmethod
    def pre_recursion(desc):
        return cleanup.cleanup(desc)

    def construct_child(self, datatype, typename=None, **kwds):
        construct = getattr(datatype, 'construct', None)
        if construct:
            return construct(self, **kwds)
        return datatype(**kwds)

    def __init__(self, *,
                 drivers, layout, maker, path, animation, controls,
                 event_queue_maxsize=EVENT_QUEUE_MAXSIZE, **kwds):
        """
        :param int event_queue_maxsize: maxsize parameter to queue.Queue.
            0 means an unbounded queue.
        """
        def post(desc):
            return self.construct_child(**desc)

        def create(root, name):
            with exception.add('Unable to create ' + name):
                return recurse.recurse(
                    root,
                    pre=None,
                    post=post,
                    python_path='bibliopixel.' + name)

        attributes.check(kwds, 'project')
        self.path = path
        layout = layout or cleanup.cleanup_layout(animation)

        self.maker = self.construct_child(**maker)
        self.drivers = [create(d, 'drivers') for d in drivers]
        with exception.add('Unable to create layout'):
            self.layout = self.construct_child(**layout)

        self.animation = create(animation, 'animation')

        eq = event_queue.EventQueue(maxsize=event_queue_maxsize)
        self.animation.event_queue = eq
        self.animation.preframe_callback = eq.get_and_run_events

        # Unfortunately, the whole animation cycle is controlled by methods on
        # the topmost animation - but we need to get called back at a certain
        # point in that animation cycle in order to run the event queue safely -
        # and animations don't know about the Project!
        #
        # So the monkey-patch above.  Ugly.  When we rewrite the
        # animation cycle, it will be freed from the topmost animation, and
        # this problem will go away.  (Also, animations should have access to
        # the Project, but that's a whole separate issue.)

        self.controls = [create(c, 'control') for c in controls]

    def start(self):
        self.layout.start()
        for c in self.controls:
            c.start(self)
        self.animation.start()

    def cleanup(self):
        self.animation.cleanup()
        for c in self.controls:
            c.cleanup()
        self.layout.cleanup_drivers()

    def run(self):
        try:
            self.start()
        finally:
            self.cleanup()


def project(*descs, root_file=None):
    """
    Make a new project, using recursion and alias resolution.

    Use this function in preference to calling Project() directly.
    """
    load.ROOT_FILE = root_file

    desc = defaults.merge(*descs)

    with load.extender(desc.get('path', '')):
        desc = recurse.recurse(desc)

    project = construct.construct(**desc)
    project.desc = desc
    return project



import math, os, shutil, tempfile, time
from . import collection
from .. util.image import gif, renderer
from .. util import colors, log


DEFAULT_RENDER = {
    'color': colors.Black,
    'pixel_width': 4,
    'pixel_height': None,
    'ellipse': False,
    'vertical': False,
    'frame': 2,
    'padding': 2,
}


class GifWriter(collection.Wrapper):
    """Write animated GIFs for each frame in the contained animation"""

    def __init__(self, *args, filename='gif_writer', render=None,
                 divide=1, frames=128, time=0, scale=1.0, gif_options=None,
                 tmp_dir=None, **kwds):
        """
        :param str filename: Base filename to write the animated GIF file

        :param dict render: Parameters to the renderer function -
            see ``bibliopixel.util.image.render.renderer``

        :param int divide: If greater than 1, only rendered one in ``divide``
            frames

        :param int frames: Number of frames to write

        :param float time: Total time to write.  If non-zero, takes precedence
            over `frames`

        :param float speed: the length of the GIF is scaled by this factor, so
            if speed=2 then a 1 second animation will become a 2 second GIF.

        :param dict gif_options: Options to
            ``bibliopixel.util.image.gif.write_animation``

        :param str tmp_dir: If set, write individual GIF frame files to this
            directory, and do not delete them when done.  For testing purposes.
        """
        super().__init__(*args, **kwds)
        render = dict(DEFAULT_RENDER, **(render or {}))
        self.render = renderer.renderer(self.layout, **render)
        self.divide = divide
        self.frames = frames
        self.time = time
        self.frame_files = []
        self.scale = scale
        self.gif_options = gif_options or {}

        filename = os.path.expanduser(os.path.abspath(filename))
        file_root = filename[:-4] if filename.endswith('.gif') else filename
        self.filename = file_root + '.gif'
        os.makedirs(os.path.dirname(self.filename), exist_ok=True)

        if tmp_dir:
            self.tmp_dir_name = tmp_dir
            shutil.rmtree(tmp_dir, ignore_errors=True)
            os.makedirs(self.tmp_dir_name, exist_ok=True)
        else:
            self.tmp_dir = tempfile.TemporaryDirectory()
            self.tmp_dir_name = self.tmp_dir.name
        self.basename = os.path.basename(file_root)
        self.finished = False
        self.stop_after_write = True

    def pre_run(self):
        super().pre_run()
        self.times = []

    def step(self, amt=1):
        super().step(amt)
        if self.finished or (self.divide >= 1 and self.cur_step % self.divide):
            return

        frame = self.cur_step / max(self.divide, 1)
        self.times.append(time.time())

        if self.time:
            elapsed = self.times[-1] - self.times[0]
            self.finished = (elapsed >= self.time)
        else:
            self.finished = (frame >= self.frames)

        if self.finished:
            self._write_gif_file()
            self.completed = True
        else:
            frame_name = '%s%04d.gif' % (self.basename, frame)
            filename = os.path.join(self.tmp_dir_name, frame_name)

            self.render().save(filename)
            self.frame_files.append(filename)

    def _write_gif_file(self):
        if not self.frame_files:
            return

        ff, self.frame_files = self.frame_files, []
        go = dict(self.gif_options)

        if 'duration' in go:
            go.pop('fps', None)
            try:
                go['duration'] *= self.scale
            except:
                go['duration'] = [g * self.scale for g in go['duration']]

        elif 'fps' in go:
            go['fps'] /= self.scale

        else:
            # Compute the list of durations from the list of times
            duration = []
            for i, time in enumerate(self.times):  # noqa: 402
                if not i:
                    previous_time = time
                else:
                    # GIFs only accept time in 0.01 second increments
                    duration.append(round(time - previous_time, 2))
                    previous_time += duration[-1]

            duration.append(duration[-1])
            go = dict(go, duration=duration)

        gif.write_animation(self.filename, ff, **go)
        self.tmp_dir = None

    def __del__(self):
        try:
            self._write_gif_file()
        except Exception as e:
            log.error('Failed to write GIF file: %s', e)



import copy, ctypes, enum
from .. server_driver import ServerDriver
from ... util import log, server_cache, udp
from . import dmx_message

UDP_PORT = 0x1936  # 6454


class ArtNet(ServerDriver):
    SERVER_CLASS = udp.Sender

    def __init__(self, *args, ip_address, **kwds):
        """
        :param dict channel_map: maps DMX channels to positions in
            the color_list
        """
        self.msg = dmx_message.dmx_message()

        address = ip_address, UDP_PORT
        super().__init__(*args, address=address, **kwds)

    def _make_buffer(self):
        return self.msg.data

    def _send_packet(self):
        # Regrettably, we need a copy here, because we can't be sure that the
        # next _render won't come before this message has been sent.
        # This is avoidable but not easily...
        self.server.send(copy.copy(self.msg))

    def _on_positions(self):
        pass



"""
An address identifies how to get or set a piece of data within a Python object
using attributes and indexing.

An address description is a string looking like:

::

    .foo.bar[32][5][baz].bang()

which would mean

"given an object target, the value ``target.foo.bar[32][5]['baz'].bang()``".

Addresses are divided into "segments".

A segment contained in brackets ``[]`` is an index (for a list) or a key (for
a dictionary) - otherwise, it's an attribute.

In the example above, the segments are ``foo``, ``bar``, ``[32]``, ``[5]``,
``[baz]`` and ``bang``; ``foo`` and ``bar`` are attributes.
``baz`` is a string index, and ``32`` and ``5`` are numeric indexes.

You can use an Address to either get or set values in a target object.

Any key that's entirely numeric is taken to be an integer index.  This is
convenient but prevents the creation of dictionaries like ``{1: 'x', '1': 'y'}``
which probably didn't want to do that anyway.
"""

from .. util import log


class Address:
    class Segment:
        def __init__(self, name):
            self.name = name

        def set(self, target, *value):
            self._set(target, (value[0] if len(value) == 1 else value))

    class Attribute(Segment):
        def get(self, target):
            return getattr(target, self.name)

        def _set(self, target, value):
            setattr(target, self.name, value)

        def __str__(self):
            return '.%s' % self.name

    class Index(Segment):
        def get(self, target):
            return target[self.name]

        def _set(self, target, value):
            target[self.name] = value

        def __str__(self):
            return '[%s]' % self.name

    class Call(Segment):
        def __init__(self):
            pass

        def get(self, target):
            return target()

        def set(self, target, *value):
            target(*value)

        def __str__(self):
            return '()'

    def __init__(self, name=None):
        if not name:
            self.segments = self.assignment = ()
            return

        self.name, *assignment = name.split('=', 1)
        self.name = self.name.strip()

        try:
            self.segments = list(_generate(self.name))
        except:
            raise ValueError('%s is not a legal address' % name)

        if not self.segments:
            raise ValueError('Empty Addresses are not allowed')

        if assignment:
            if not self.segments:
                raise ValueError('Cannot assign to an empty address')
            if isinstance(self.segments[-1], Address.Call):
                raise ValueError('Cannot assign to a call operation')
            assignment = [s.strip() for s in assignment[0].split(',')]
            assignment = [int(s) if s.isnumeric() else s for s in assignment]
            self.assignment = tuple(assignment)
        else:
            self.assignment = ()

    def __bool__(self):
        return bool(self.segments)

    def __str__(self):
        return self.name

    @staticmethod
    def _get(target, address):
        for a in address:
            target = a.get(target)
        return target

    def get(self, target):
        return self._get(target, self.segments)

    def set(self, target, *values):
        *first, last = self.segments
        parent = self._get(target, first)
        last.set(parent, *(self.assignment + values))


def _generate(s):
    def extract_calls(p):
        # Extract () pairs from start and finish of a string
        before, after = [], []
        while p.startswith('()'):
            before.append(Address.Call())
            p = p[2:]

        while p.endswith('()'):
            after.append(Address.Call())
            p = p[:-2]

        return before, p, after

    # Split on dots, then use [ and ] to split out indices
    s = s.strip()
    if s.startswith('.'):
        s = s[1:]
    if s.endswith('.'):
        raise ValueError

    for i, part in enumerate(s.split('.')):
        if not part:
            raise ValueError

        head, *rest = part.split('[')

        # If we had e.g. 'xxx()()[yyy]()[zzz]()()'
        # Now we have first='xxx()' and rest = 'yyy]()', 'zzz]()()'
        # They might have written: ()xxx by mistake
        before, head, after = extract_calls(head)

        if before:
            # A call () is only allowed to start the first segment -
            # for example, an address like a.() is forbidden.
            if i or head:
                raise ValueError
            yield from before

        elif head:
            yield Address.Attribute(head)
            yield from after

        elif i:
            # An index [] is only allowed to start the first segment -
            # for example, an address like a.[2] is forbidden.
            raise ValueError

        for r in rest:
            before, r, after = extract_calls(r)
            if before:
                # A segment cannot contain a () - they must all be at top level.
                raise ValueError

            # A segment must have exactly one ']', exactly at the end.
            index, between = r.split(']', 1)
            if between:
                raise ValueError

            index = int(index) if index.isdigit() else index
            yield Address.Index(index)
            yield from after



import webbrowser

DEFAULT_URL = 'http://simpixel.io'


def open_simpixel(url=None):
    url = url or DEFAULT_URL

    if not url.startswith('no'):
        webbrowser.open(url, new=0, autoraise=True)



import os
from .. project.importer import import_module

COMMANDS = (
    'all_pixel', 'all_pixel_test', 'clear_cache', 'color', 'demo',
    'devices', 'info', 'load', 'list', 'monitor', 'remove', 'reset', 'run',
    'save', 'set', 'show', 'update')

MODULES = {c: import_module('bibliopixel.main.' + c) for c in COMMANDS}

RST_HELP = """\
## ``bp`` - The BiblioPixel Project Runner

``bp`` is a command-line script installed with Bibliopixel.  It can run
projects and demos, configure hardware, save and load defaults, and more.

``bp`` takes one of {0} *commands*:
  {1}
  {2}

Some commands take *arguments* - often names of files.

### Examples of ``bp`` commands and arguments

``````
bp info
bp demo circle
bp run foo.json bar.json
bp foo.json bar.json      # run is the default command
``````

The commands are summarized below, and you can get more detailed help about
each one by typing:

``````
bp <command> --help
``````

or

``````
bp <command> -h
``````

### The ``bp`` command line

``bp`` commands have *flags* that control how ``bp`` runs its commands.

Examples of flags are ``-v``, ``--verbose``, ``--defaults``, ``-d``,
``--simpixel``.  Flags either start with ``-``, for one letter flags, or ``--``
for multi-letter flags.

Sometimes two flags have the same meaning, like ``--simpixel`` and ``-s``
or ``--verbose`` and ``-v``.

Some flags take an argument, like ``--dim=160`` or ``--loglevel=frame``.
Other flags do not, like ``--verbose`` or ``-v``.

A ``bp`` command line can optionally include a command, a list of arguments,
and a list of flags:

``````
bp [<command>] [<argument> ...] [<flag> ...]
``````

### Examples of ``bp`` command lines

``````
bp info --help
bp run -h
bp info
bp color "indian red 4"
bp foo.json bar.json -v
bp foo.json bar.json -v --numbers=float --dump --defaults=default.json
``````
""".format(len(COMMANDS),
           ', '.join(COMMANDS[0:8]),
           ', '.join(COMMANDS[8:]))

SEP = '``````'


def un_md(s):
    return s.replace(SEP, '').replace('``', '`')


HELP = un_md(RST_HELP)

COMMAND_HELP_PREFIX = """
### APPENDIX: ``bp <command> --help`` for each command
"""


def help_text():
    def helper():
        yield RST_HELP
        yield COMMAND_HELP_PREFIX

        for i, c in enumerate(COMMANDS):
            if i:
                yield ''
                yield '______________________________________________'
                yield ''

            module = MODULES[c]
            yield '``bp %s``' % c
            yield ''
            yield '*%s*' % module.__doc__.strip()
            yield ''
            yield getattr(module, 'DESCRIPTION', '')

    return '\n'.join(helper()) + '\n'


def extract_help(filename):
    with open(filename, 'w') as fp:
        fp.write(help_text())



import getpass, platform, sys
from .. util import log
from . control import ExtractedControl

# See https://stackoverflow.com/questions/42603000
DARWIN_ROOT_WARNING = """
In MacOS, pynput must to be running as root in order to get keystrokes.

Try running your program like this:

    sudo %s <your commands here>
"""

INSTALL_ERROR = """
Please install the pynput library with

    $ pip install pynput

"""


def keyname(key):
    return getattr(key, 'name', None) or getattr(key, 'char')


class Keyboard(ExtractedControl):
    EXTRACTOR = {
        'keys_by_type': {
            'press': ['type', 'key'],
            'release': ['type', 'key'],
        },

        'normalizers': {
            'key': keyname,
        },
    }

    def _press(self, key):
        self.receive({'type': 'press', 'key': key})

    def _release(self, key):
        self.receive({'type': 'release', 'key': key})

    def _make_thread(self):
        try:
            import pynput
        except Exception as e:
            e.args = (INSTALL_ERROR,) + e.args
            raise

        if platform.platform().startswith('Darwin'):
            if getpass.getuser() != 'root':
                log.warning(DARWIN_ROOT_WARNING, sys.argv[0])

        return pynput.keyboard.Listener(self._press, self._release)


class KeyboardTester:
    def __setattr__(self, k, v):
        super().__setattr__(k, v)
        log.printer(k, '=', v)


def main():
    keyboard = Keyboard(routing={'press': 'press', 'release': 'release'})
    keyboard.start(KeyboardTester())
    keyboard.thread.join()


if __name__ == '__main__':
    main()



import collections, functools, sys, threading
from . extractor import Extractor
from .. project import construct, importer
from .. util.log_errors import LogErrors
from .. util import flatten, log, json
from . routing import ActionList, Routing


class Control:
    DEFAULT = {'datatype': ActionList}

    def __init__(self, routing=None, default=None, max_errors=16,
                 python_path='bibliopixel.control', verbose=False,
                 pre_routing=None):
        """
        :param Address pre_routing: This Address is set with with the message
            after the message is received and converted, but before it is
            routed.
        """
        self.verbose = verbose
        self.receive = LogErrors(self._receive, max_errors)
        default = dict(self.DEFAULT, **(default or {}))
        self.pre_routing = ActionList(pre_routing)
        self.routing = Routing(routing or {}, default or {}, python_path)
        self.running = False

    def start(self, project):
        if self.verbose:
            log.info('Starting %s', self)

        self.project = project  # Can I now delete this?
        self.pre_routing.set_target(project)
        self.routing.set_target(project)
        self.running = True
        self.thread = self._make_thread()
        self.thread.start()

    def cleanup(self):
        self.stop()

    def stop(self):
        self.running = False

    def loop(self):
        for msg in self:
            self.receive(msg)
            if not self.running:
                return

    def _receive(self, msg):
        """
        Receive a message from the input source and perhaps raise an Exception.
        """
        msg = self._convert(msg)

        str_msg = self.verbose and self._msg_to_str(msg)
        if self.verbose and log.is_debug():
            log.debug('Message %s', str_msg)

        if self.pre_routing:
            self.pre_routing.receive(msg)

        receiver, msg = self.routing.receive(msg)
        if receiver:
            receiver.receive(msg)
            if self.verbose:
                log.info('Routed message %s to %s', str_msg, receiver)

    def _convert(self, msg):
        """
        Convert the message to a Control-specific format
        """
        raise NotImplementedError

    def __iter__(self):
        """Should yield a sequence of messages from the input source."""
        raise NotImplementedError

    def _make_thread(self):
        """
        Returns the thread that run the loop for this control source.

        If the underlying source has its own thread or process, override this
        method.  The object returned needs to have two methods:  ``start()``
        and ``stop()``.
        """
        return threading.Thread(target=self.loop, daemon=True)

    def _msg_to_str(self, msg):
        if msg is None:
            return '(None)'
        return '.'.join(msg.values()) or '.'

    def __bool__(self):
        return bool(self.routing or self.pre_routing)


class ExtractedControl(Control):
    EXTRACTOR = {}

    def __init__(self, extractor=None, **kwds):
        super().__init__(**kwds)
        extractor = dict(self.EXTRACTOR, **(extractor or {}))
        self._convert = Extractor(**extractor).extract



from .. util import class_name, log


class LogErrors:
    """
    Wraps a function call to catch and report exceptions.
    """

    def __init__(self, function, max_errors=-1):
        """
        :param function: the function to wrap
        :param int max_errors: if ``max_errors`` is non-zero, then only the
                               first ``max_errors`` error messages are printed
        """
        self.function = function
        self.max_errors = max_errors
        self.errors = 0

    def __call__(self, *args, **kwds):
        """
        Calls ``self.function`` with the given arguments and keywords, and
        returns its value - or if the call throws an exception, returns None.

        If is ``self.max_errors`` is `0`, all the exceptions are reported,
        otherwise just the first ``self.max_errors`` are.
        """
        try:
            return self.function(*args, **kwds)
        except Exception as e:
            args = (class_name.class_name(e),) + e.args

        self.errors += 1
        if self.max_errors < 0 or self.errors <= self.max_errors:
            log.error(str(args))

        elif self.errors == self.max_errors + 1:
            log.error('Exceeded max_errors of %d', self.max_errors)



"""
Run specified project from file or URL.
"""

import os, sys, time, traceback
from . import common_flags, simpixel
from .. util import json, log
from .. animation import Animation
from .. project import load

RUN_ERROR = """When reading file {filename}:
{desc}
"""

FAILURE_ERROR = """\
{count} project{s} failed
____________________
"""


def _load_py(filename):
    module = load.module(filename)
    project = module.get('PROJECT', {})

    if 'animation' not in project:
        no_extension = os.path.splitext(filename)[0]
        module_name = os.path.split(no_extension)[-1]
        names = module.keys()
        try:
            animation_name = load.guess_name(names, module_name, filename)
        except:
            if 'Animation' not in names:
                raise ValueError('Cannot deduce animation in file ' + filename)
            animation_name = 'Animation'
        animation = module[animation_name]
        if not callable(animation):
            raise ValueError('Animation "%s" in file "%s" is not callable'
                             % (animation_name, filename))

        project['animation'] = {'datatype': animation}

    return project


def _dump(args, desc):
    if not isinstance(desc, str):
        desc = json.yaml.dump(desc) if args.yaml else json.dumps(desc)
    return desc.strip()


def _get_projects(args):
    projects, failed = [], []

    for filename in args.name:
        saved_path = sys.path[:]
        desc = '(not loaded)'
        try:
            if filename.endswith('.py'):
                desc = _load_py(filename)
                root_file = None
            else:
                desc = load.data(filename, False)
                desc = json.loads(desc, filename)
                root_file = os.path.abspath(filename)

            project = common_flags.make_project(args, desc, root_file)
            projects.append(project)
            if args.dump:
                print(_dump(args, project.desc))

        except FileNotFoundError as e:
            failed.append(('%s: %s' % (e.strerror, e.filename), ()))

        except Exception as exception:
            if args.verbose or filename.endswith('.py'):
                raise
            eargs = exception.args
            if args.verbose:
                exception = traceback.format_exc()
            desc = desc and _dump(args, desc)
            msg = RUN_ERROR.format(**locals())
            failed.append((msg, eargs))

        finally:
            sys.path[:] = saved_path

    if not failed:
        return projects

    log.error(FAILURE_ERROR.format(
        count=len(failed), s='' if len(failed) == 1 else 's'))
    for msg, args in failed:
        if args:
            log.error(msg + '\n' + '\n'.join(str(a) for a in args) + '\n')
        else:
            log.error(msg + '\n')
    raise ValueError('Run aborted')


def _run_projects(projects, args):
    needs_pause = False
    assert len(projects) == len(args.name)
    for project, name in zip(projects, args.name):
        log.debug('Running file %s', name)
        if needs_pause:
            args.pause and time.sleep(float(args.pause))
        else:
            needs_pause = True

        try:
            project.run()

        except KeyboardInterrupt:
            if not args.ignore_exceptions:
                raise
            log.warning('\nKeyboardInterrupt terminated project.')
            needs_pause = False

        except Exception as e:
            if not args.ignore_exceptions:
                raise
            log.error('Exception %s', e)
            traceback.print_exc()


def run(args):
    Animation.FAIL_ON_EXCEPTION = args.fail_on_exception

    args.name = args.name or ['']
    projects = _get_projects(args)

    if args.dry_run:
        print('(dry run - nothing executed)')
        return
    if args.simpixel:
        simpixel.open_simpixel(args.simpixel)
    elif args.s:
        simpixel.open_simpixel()

    _run_projects(projects, args)


def set_parser(parser):
    parser.set_defaults(run=run)
    common_flags.add_project_flags(parser)

    parser.add_argument(
        'name', nargs='*',
        help='Path project files - can be a URL or file system location')

    parser.add_argument(
        '-j', '--json', action='store_true',
        help='Enter JSON directly as a command line argument.')



import os, traceback
from .. project import aliases, construct, load, project
from . import animation
from .. util import json, log


class Collection(animation.Animation):
    """
    A ``Collection`` is a list of ``Animation``s
    """

    @staticmethod
    def pre_recursion(desc):
        animations = []

        for a in desc['animations']:
            loaded = load.load_if_filename(a)
            if loaded:
                animation = loaded.get('animation', {})
                run = loaded.get('run', {})

            elif callable(a) or isinstance(a, str) or 'animation' not in a:
                animation = a
                run = {}

            else:
                animation = a.pop('animation', {})
                run = a.pop('run', {})
                if a:
                    raise ValueError(
                        'Extra fields in animation: ' + ', '.join(a))

            animation = construct.to_type_constructor(
                animation, 'bibliopixel.animation')

            run.update(animation.get('run', {}))
            animation['run'] = run

            # Children without fps or sleep_time get it from their parents.
            if not ('fps' in run or 'sleep_time' in run):
                if 'fps' in desc['run']:
                    run.update(fps=desc['run']['fps'])
                elif 'sleep_time' in desc['run']:
                    run.update(sleep_time=desc['run']['sleep_time'])

            animations.append(animation)

        desc['animations'] = animations
        return desc

    CHILDREN = 'animations',

    def __init__(self, layout, animations=None, **kwds):
        super().__init__(layout, **kwds)
        self.animations = animations or []
        self.internal_delay = 0  # never wait

    # Override to handle all the animations
    def cleanup(self, clean_layout=True):
        self.state = animation.STATE.canceled
        for a in self.animations:
            a.cleanup()
        super().cleanup(clean_layout)

    def add_animation(self, anim, **kwds):
        from .. util import deprecated
        deprecated.deprecated('Collection.add_animation')

        anim._set_runner(kwds)
        self.animations.append(anim)

    def pre_run(self):
        for a in self.animations:
            a.pre_run()


class Indexed(Collection):
    """
    An ``Indexed`` is a :py:class:`bibliopixel.animation.Collection`
    which keeps track of the current animation through an index into the list of
    animations, and gets a callback after that index changes.

    An ``Indexed`` has two properties.

    ``index`` is a mutable property indexing the current animation in the list
    of animations.

    ``current_animation`` returns the animation at position ``index`` or None
    if it is out of the bounds of the ``Collection``.
    """
    _index = -1

    @property
    def index(self):
        """
        :returns int: index of the current animation within the Collection.
        """
        return self._index

    @index.setter
    def index(self, index):
        self._index, old_index = index, self._index
        self._on_index(old_index)

    def _on_index(self, old_index):
        """
        Override this method to get called right after ``self.index`` is set.

        :param int old_index: the previous index, before it was changed.
        """
        if self.current_animation:
            log.debug('%s: %s',
                      self.__class__.__name__, self.current_animation.title)
            self.frames = self.current_animation.generate_frames(False)

    @property
    def current_animation(self):
        """
        :returns: the selected animation based on self.index, or None if
            self.index is out of bounds
        """
        if 0 <= self._index < len(self.animations):
            return self.animations[self._index]


class Parallel(Collection):
    def __init__(self, *args, levels, **kwds):
        super()(*args, **kwds)

        # Give each animation a unique, mutable layout so they can
        # run independently.
        for a in self.animations:
            a.layout = a.layout.mutable_copy()


class Wrapper(Indexed):
    # TODO: No unit tests cover any of this.
    @staticmethod
    def pre_recursion(desc):
        if 'animations' in desc:
            raise ValueError('Cannot specify animations in a Wrapper')
        desc['animations'] = [desc.pop('animation')]
        return Collection.pre_recursion(desc)

    def pre_run(self):
        super().pre_run()
        self.index = 0

    @property
    def animation(self):
        # This is used elsewhere in client code.  TODO: pick one or the other!
        return self.current_animation



"""
Monitor a control source
"""

import sys
from .. project import importer
from .. util import log


def run(args):
    try:
        tc = importer.import_module(args.control[0], 'bibliopixel.control')
    except:
        log.error('Do not understand control "%s"', args.control[0])
        raise

    try:
        main = tc.main
    except:
        log.error('No main() function for control "%s"', args.control)
        raise

    main()


def set_parser(parser):
    parser.add_argument(
        'control', nargs=1,
        help='Name of a control to monitor')

    parser.set_defaults(run=run)



from . address import Address
from . receiver import Receiver


class QueuedAddress(Receiver):
    """
    A `QueuedAddress` is a `Receiver` which sets `Address`es, perhaps using an
    an optional `EventQueue`.

    When the `set_target` method is called, `QueuedAddress` searches
    down through the address and stores the most recent `event_queue`
    method it finds.

    """
    def __init__(self, address=None):
        self.address = Address(address)

    def set_target(self, target):
        self.last_segment = self.address.segments[-1]
        self.queue = None
        self.subtarget = target

        for segment in self.address.segments:
            self.queue = getattr(self.subtarget, 'event_queue', self.queue)
            if segment is not self.last_segment:
                self.subtarget = segment.get(self.subtarget)

    def receive(self, msg):
        """
        Receives a message, and either sets it immediately, or puts it on the
        event queue if there is one.

        """
        if self.queue:
            self.queue.put_event(self._set, msg)
        else:
            self._set(msg)

    def __bool__(self):
        return bool(self.address)

    def __str__(self):
        return str(self.address)

    def get(self):
        return self.last_segment.get(self.subtarget)

    def _set(self, values):
        args = self.address.assignment + values
        self.last_segment.set(self.subtarget, *args)



import fractions
from . import control
from .. util import log

try:
    import mido
    MESSAGE_TYPES = set(s['type'] for s in mido.messages.specs.SPECS)

except:
    mido = None
    MESSAGE_TYPES = set()


class Midi(control.ExtractedControl):
    EXTRACTOR = {
        # There are more mido message types that we haven't used yet.
        'keys_by_type': {
            'aftertouch': (
                'port', 'channel', 'type', 'value'),
            'control_change': (
                'port', 'channel', 'type', 'control', 'value'),
            'note_off': (
                'port', 'channel', 'type', 'note', 'velocity'),
            'note_on': (
                'port', 'channel', 'type', 'note', 'velocity'),
            'pitchwheel': (
                'port', 'channel', 'type', 'pitch'),
            'program_change': (
                'port', 'channel', 'type', 'program'),
        },

        # Some numeric fields (channel, control, program and note) don't get
        # normalized because they are basically names.
        'normalizers': {
            'pitch': lambda x: fractions.Fraction(x - 8192) / 8192,
            'value': lambda x: fractions.Fraction(x) / 127,
            'velocity': lambda x: fractions.Fraction(x) / 127,
        },

        'omit': ('port', 'channel'),
    }

    def __init__(self, use_note_off=False, **kwds):
        super().__init__(**kwds)
        self.use_note_off = use_note_off

    def __iter__(self):
        ports = [mido.open_input(i) for i in mido.get_input_names()]

        if not ports:
            log.error('control.midi: no MIDI ports found')
            return

        log.info('Starting to listen for MIDI')
        for port, msg in mido.ports.MultiPort(ports, yield_ports=True):
            mdict = dict(vars(msg), port=port)
            if self.use_note_off or msg.type != 'note_off':
                yield mdict
            else:
                yield dict(mdict, type='note_on', velocity=0)


def _print_midi(*args, **kwds):
    print('MIDI:', *args)


def main():
    midi = Midi(pre_routing='()')
    midi.start(_print_midi)
    midi.thread.join()


if __name__ == '__main__':
    main()



from . animation import Animation
from .. layout import Circle as CircleLayout


class Circle(Animation):

    def __init__(self, layout, **kwds):
        super().__init__(layout, **kwds)

        if not isinstance(layout, CircleLayout):
            raise RuntimeError(
                'Must use %s with Circle animations, not %s' %
                CircleLayout, type(layout))

        self.rings = layout.rings
        self.ringCount = layout.ringCount
        self.lastRing = layout.lastRing
        self.ringSteps = layout.ringSteps


from .. util import deprecated
if deprecated.allowed():
    BaseCircleAnim = Circle



import contextlib, threading, time
from . runner import Runner, STATE
from .. util import deprecated, log
from .. util.threads.animation_threading import AnimationThreading
from .. project import attributes, fields


class Animation(object):
    free_run = False
    pre_recursion = fields.default_converter
    FAIL_ON_EXCEPTION = False

    @classmethod
    def construct(cls, project, *, run=None, name=None, data=None, **desc):
        """
        Construct an animation, set the runner, and add in the two
        "reserved fields" `name` and `data`.
        """
        try:
            a = cls(project.layout, **desc)
            a._set_runner(run or {})
        except Exception as e:
            if cls.FAIL_ON_EXCEPTION:
                raise
            from . failed import Failed
            a = Failed(project.layout, desc, e)

        a.name = name
        a.data = data
        return a

    def __init__(self, layout, *, preclear=True, **kwds):
        attributes.set_reserved(self, 'animation', **kwds)
        self.layout = layout
        self.internal_delay = None
        self.on_completion = None
        self.state = STATE.ready
        self.preclear = preclear

    @property
    def _led(self):
        """Many BiblioPixelAnimations use the "protected" variable _led."""
        return self.layout

    @_led.setter
    def _led(self, layout):
        self.layout = layout

    @property
    def color_list(self):
        return self.layout.color_list

    @color_list.setter
    def color_list(self, cl):
        self.layout.color_list[:] = cl

    @property
    def completed(self):
        """Many BiblioPixelAnimations use the old `completed` variable."""
        return self.state == STATE.complete

    @completed.setter
    def completed(self, state):
        self.state = STATE.complete if state else STATE.running

    def pre_run(self):
        pass

    @property
    def title(self):
        return self.__class__.__name__

    def step(self, amt=1):
        pass

    def cleanup(self, clean_layout=True):
        self.threading.cleanup()
        # Some cases we may not want to clear the screen
        # Like with the remote, it would flash between anims
        if clean_layout:
            self.layout.cleanup()

    def preframe_callback(self):
        """
        preframe_callback is called right before the start of a frame rendering
        pass.

        The ``Project`` writes over this method, for the top-level animation
        only, in order to drain its event_queue at a moment where no rendering
        is happening to avoid race conditions.
        """
        pass

    def start(self):
        self.threading.start()

    def stop(self):
        self.threading.stop_event.set()

    def run_all_frames(self):
        for i in self.generate_frames():
            pass

    def generate_frames(self, clean_layout=True):
        with self._run_context(clean_layout):
            while self.state == STATE.running:
                self._run_one_frame()
                yield

    def run(self, **kwds):
        deprecated.deprecated('BaseAnimation.run')
        self._set_runner(kwds)
        self.start()

    def _check_delay(self):
        if self.free_run:
            self.sleep_time = None
        elif self.internal_delay:
            self.sleep_time = self.internal_delay
        else:
            self.sleep_time = self.runner.sleep_time
        self.layout.animation_sleep_time = self.sleep_time or 0

    def _run_one_frame(self):
        timestamps = [time.time()]

        self.preframe_callback()
        timestamps = [time.time()]

        self.step(self.runner.amt)
        timestamps.append(time.time())

        self.layout.frame_render_time = timestamps[1] - timestamps[0]
        self.layout.push_to_driver()

        timestamps.append(time.time())
        _report_framerate(timestamps)

        self.cur_step += 1
        if self.state == STATE.complete and self.runner.max_cycles > 0:
            if self.cycle_count < self.runner.max_cycles - 1:
                self.cycle_count += 1
                self.state = STATE.running

        self.threading.wait(self.sleep_time, timestamps)
        if self.threading.stop_event.isSet():
            self.state = STATE.canceled
        else:
            self.state = self.runner.compute_state(self.cur_step, self.state)

    @contextlib.contextmanager
    def _run_context(self, clean_layout=True):
        self.state = STATE.running
        self.runner.run_start_time = time.time()
        self.threading.stop_event.clear()

        if deprecated.allowed():
            self._step = 0
        self.cur_step = 0
        self.cycle_count = 0

        self._check_delay()
        self.preclear and self.layout.all_off()
        self.pre_run()

        try:
            yield
        finally:
            self.cleanup(clean_layout)

        self.on_completion and self.on_completion(self.state)
        self.state = STATE.ready

    def _set_runner(self, runner):
        self.runner = Runner(**(runner or {}))
        self.threading = AnimationThreading(self.runner, self.run_all_frames)


if deprecated.allowed():
    BaseAnimation = Animation


def _report_framerate(timestamps):
    total_time = timestamps[-1] - timestamps[0]
    fps = int(1.0 / max(total_time, 0.001))
    log.frame("%dms/%dfps / Frame: %dms / Update: %dms",
              1000 * total_time,
              fps,
              1000 * (timestamps[1] - timestamps[0]),
              1000 * (timestamps[2] - timestamps[1]))



import ctypes

DMX_LENGTH = 512
ARTNET_DMX = 0x5000
ARTNET_VERSION = 14
NAME = 'Art-Net\x00'
MAX_NET = 0xFF
MAX_SUBNET = 0xF
MAX_UNIVERSE = 0xF


def dmx_message(data=None, length=None, net=0, subnet=0, universe=0,
                sequence=1):
    if length is None:
        if data is None:
            length = DMX_LENGTH
        else:
            length = len(data)

    assert length % 2 == 0, 'artnet only takes messages of even length'
    assert 0 <= length <= DMX_LENGTH
    assert 0 <= sequence <= DMX_LENGTH
    assert 0 <= net <= MAX_NET
    assert 0 <= subnet <= MAX_SUBNET
    assert 0 <= universe <= MAX_UNIVERSE

    Char, Int8, Int16 = ctypes.c_char, ctypes.c_ubyte, ctypes.c_ushort

    class ArtnetDMXMessage(ctypes.Structure):
        # http://artisticlicence.com/WebSiteMaster/User%20Guides/art-net.pdf p47
        _fields_ = [
            ('id', Char * 8),
            ('opCode', Int16),
            ('protVerHi', Int8),
            ('protVerLo', Int8),
            ('sequence', Int8),
            ('physical', Int8),
            ('subUni', Int8),
            ('net', Int8),
            ('lengthHi', Int8),
            ('length', Int8),
            ('data', Int8 * length),  # At position 18
        ]

    # http://artisticlicence.com/WebSiteMaster/User%20Guides/art-net.pdf p5
    subUni = (subnet << 4) + universe
    hi, lo = divmod(length, 256)

    msg = ArtnetDMXMessage(
        id=NAME.encode(),
        opCode=ARTNET_DMX,
        protVerLo=ARTNET_VERSION,
        sequence=sequence,
        net=net,
        subUni=subUni,
        lengthHi=hi,
        length=lo)

    if data is not None:
        msg.data[:] = data

    return msg


Message = type(dmx_message())



import functools, threading, queue


class Loop(threading.Thread):
    def __init__(self, loop=None, daemon=True, **kwds):
        super().__init__(daemon=daemon, **kwds)
        self.running = True
        self.loop = loop or self.loop

    def run(self):
        while self.running:
            self.loop()

    def stop(self):
        self.running = False

    def loop(self):
        pass


class Loops(object):
    def __init__(self, loops):
        self.loops = list(loops)

    def start(self):
        map(lambda x: x.start(), self.loops)

    def stop(self):
        map(lambda x: x.stop(), self.loops)


class QueueHandler(Loop):
    def __init__(self, timeout=0.1, send=None, **kwds):
        super().__init__(**kwds)
        self.timeout = timeout
        self.queue = queue.Queue
        self.send = send or self.send

    def loop(self):
        try:
            msg = self.queue.get(timeout=self.timeout)
        except queue.Empty:
            pass
        else:
            self.send(msg)

    def send(self, msg):
        pass
